@@ -6,14 +6,22 @@ use Apache::Constants qw(OK);
 @ISA = qw(Exporter);
 @EXPORT_OK = qw(exit warn);
 $VERSION = "1.11";
+$Apache::CRLF = "\015\012";
 
 bootstrap Apache $VERSION;
 
+*Apache::warn = \&Apache::log_error;
+
+sub module {
+    my($self, $module) = @_;
+    eval "require $module;";
+}
+
 sub parse_args {
     my($wantarray,$string) = @_;
     return unless defined $string and $string;
     if(defined $wantarray and $wantarray) {
-	return map { Apache::unescape_url($_) } split /[=&]/, $string;
+	return map { Apache::unescape_url_info($_) } split /[=&]/, $string;
     }
     $string;
 }
@@ -38,6 +46,7 @@ sub cgi_var {
     return $val;
 }
 
+*READ = \&read;
 sub read {
     my($r, $bufsiz) = @_[0,2];
     my($nrd, $buf, $total);
@@ -47,8 +56,8 @@ sub read {
     $r->hard_timeout("Apache->read");
 
     while($bufsiz) {
-	$nrd = $r->read_client_block($buf, $bufsiz);
-	if($nrd > 0) {
+	$nrd = $r->read_client_block($buf, $bufsiz) || 0;
+	if(defined $nrd and $nrd > 0) {
 	    $bufsiz -= $nrd;
 	    $total += $nrd;
 	    $_[1] .= $buf;
@@ -63,6 +72,17 @@ sub read {
     return $total;
 }
 
+sub GETC { my $c; shift->read($c,1); $c; }
+
+#shouldn't use <STDIN> anyhow, but we'll be nice
+sub READLINE { 
+    my $r = shift;
+    my $line; 
+    $r->read($line, $r->header_in('Content-length'));
+    $line;
+}
+
+*PRINT = \&print;
 sub print {
     my($r) = shift;
     $r->hard_timeout("Apache->print");
@@ -139,13 +159,15 @@ sub as_string {
     join "\n", @retval, "";
 }
 
-sub untaint {
-    my $self = shift;
-    for (@_) {
-	$_ =~ /^(.*)$/s; $_ = $1;
-    }
+sub TIEHANDLE {
+    my($class, $r) = @_;
+    $r ||= Apache->request;
 }
 
+#some backwards-compatible stuff
+*Apache::TieHandle::TIEHANDLE = \&TIEHANDLE;
+*CGI::Apache::exit = \&Apache::exit;
+
 1;
 
 __END__
@@ -189,11 +211,25 @@ by passing an object reference to C<Apache->request($r)>.
 If handlers use modules such as C<Apache::CGI> that need to access
 L<Apache->request>, they too should do this (e.g. Apache::Status).
 
+=item $r->as_string
+
+Returns a string representation of the request object.
+
 =item $r->main
 
 If the current request is a sub-request, this method returns a blessed 
 reference to the main request structure.
 
+=item $r->prev
+
+This method returns a blessed reference to the previous (internal) request
+structure.
+
+=item $r->next
+
+This method returns a blessed reference to the next (internal) request
+structure.
+
 =item $r->is_main
 
 Returns true if the current request object is for the main request.
@@ -685,7 +721,13 @@ An alias for Apache->log_error.
 
 =item Apache::unescape_url($string)
 
-Handy function for unescapes.
+Handy function for unescapes.  Use this one for filenames/paths.
+Use unescape_url_info for the result of submitted form data.
+
+=item Apache::unescape_url_info($string)
+
+Handy function for unescapes submitted form data.
+In opposite to unescape_url it translates the plus sign to space.
 
 =item Apache::perl_hook($hook)
 
